section \<open>Congruence Closure (2 persons)\<close>

text \<open>We consider a set ground equations GE such as

  \<^item> f(g(a)) = h(b)
  \<^item> f(b) = b
  \<^item> g(a) = b
   
  and are interested in the question whether a particular equation is implied GE.
  For instance the sequence of equality-steps 

  \<^item> f(h(b)) = f(f(g(a))) = f(f(b)) = f(b)     

  proves that f(h(b)) = f(b) follows from E.
  
  Whereas it is easy to validate a given sequence of equality-steps, the problem
  is to detect whether such a sequence exists for a given equation. To this end,
  the congruence closure algorithm has been developed which should be partially verified in 
  this project.

  Basic knowledge of term rewriting is helpful for this project.
  The describtion of the algorithm is based on 
  \<open>Franz Baader and Tobias Nipkow, Term Rewriting and All That, Chapter 4.3\<close>.
 \<close>

theory Project_Congruence_Closure
  imports 
    Main
begin

subsection \<open>Definition of Algorithm\<close>

text \<open>We start by definining ground terms where the type of symbols are just strings.\<close>
type_synonym symbol = string

datatype trm = Fun symbol "trm list" 

type_synonym eqs = "(trm \<times> trm)set" 

text \<open>Define the set of subterms of a term, e.g., the subterms of f(g(a),b) would be
  \<open>{f(g(a),b), g(a), a, b}\<close>.\<close>

fun subt :: "trm \<Rightarrow> trm set" where
  "subt (Fun f ts) = undefined" 

text \<open>Prove two useful lemmas about subterms.\<close>
lemma self_subt: "u \<in> subt u" sorry

lemma subt_trans: "s \<in> subt t \<Longrightarrow> t \<in> subt u \<Longrightarrow> s \<in> subt u" sorry

text \<open>For a set of ground-equalities, the congruence closure algorithm is in 
  particular interested in all subterms that occur in the equalities.\<close>
definition subt_eqs where "subt_eqs GE = \<Union> ((\<lambda> (l,r). subt l \<union> subt r) ` GE)" 


text \<open>From now on fix a specific set of ground-equalities GE.\<close>
context 
  fixes GE :: eqs
begin

text \<open>Define an equality step where one can either replace one side of an equation in GE by the other side 
  (a root-step), or where one can apply a step in a context.\<close>

inductive_set estep :: "trm rel" where
  root: "undefined \<Longrightarrow> undefined \<in> estep" 
| ctxt: "(s,t) \<in> estep \<Longrightarrow> (Fun f (before @ s # after), Fun f (before @ t # after)) \<in> estep" 

text \<open>The other important definition is the Cong-operation which given a set of equalities derives
  new equalities of these by reflexivity, symmetry, transitivity or context.\<close>
inductive_set Cong :: "eqs \<Rightarrow> eqs" for E where 
  C_keep: "eq \<in> E \<Longrightarrow> eq \<in> Cong E" 
| C_refl: "(t,t) \<in> Cong E" 
| C_sym: "(s,t) \<in> E \<Longrightarrow> (t,s) \<in> Cong E" 
| C_trans: "(s,t) \<in> E \<Longrightarrow> (t,u) \<in> E \<Longrightarrow> (s,u) \<in> Cong E" 
| C_cong: "length ss = length ts \<Longrightarrow> (\<forall> i < length ts. (ss ! i, ts ! i) \<in> E) \<Longrightarrow> (Fun f ss, Fun f ts) \<in> Cong E" 

text \<open>Let us now fix to terms s and t where we are interested in whether GE implies s = t.\<close>
context 
  fixes s t :: trm
begin

text \<open>In the congruence closure algorithm one only is interested in equalities of terms in S.\<close>

definition S where "S = subt s \<union> subt t \<union> subt_eqs GE" 

definition CongS where "CongS E = Cong E \<inter> (S \<times> S)" 

text \<open>CCA defines the equalities that are obtained in the i-th iteration of the congruence closure 
  algorithm, which iteratively applies the @{const CongS} operation starting from @{term GE}.\<close>

definition CCA where "CCA i = (CongS ^^ i) GE"

text \<open>Prove the following simple inclusions.\<close>
lemma GE_S: "GE \<subseteq> S \<times> S" sorry

lemma GE_CCA: "GE \<subseteq> CCA i" sorry


subsection \<open>Completeness of CCA\<close>

text \<open>The crucial result of the congruence closure algorithm is given in the following 
  lemma on the completeness of the algorithm: if 
  the algorithm has stabilized in the i-th iteration, then all equations in 
  @{term "S \<times> S"} that can be derived with arbitrary many steps are also contained 
  in the equalities of CCA.\<close>

lemma esteps_imp_CCA: assumes "CongS (CCA i) = CCA i" 
  shows "(u,v) \<in> estep^* \<inter> (S \<times> S) \<longrightarrow> (u,v) \<in> CCA i" 
proof 
  text \<open>The proof is by induction on the number of steps and then by the size of the starting
    term @{term u}. This is expressed as follows in Isabelle.\<close>
  assume "(u,v) \<in> estep^* \<inter> (S \<times> S)" 
  then obtain n where *: "u \<in> S" "v \<in> S" "(u,v) \<in> estep^^n"  
    by (auto simp: rtrancl_power)
  obtain m where "m = (n,size u)" by auto
  with * show "(u,v) \<in> CCA i" 
  proof (induction m arbitrary: u v n rule: wf_induct[OF wf_measures[of "[fst,snd]"]])
    case (1 m u v n)

    text \<open>For handling the induction, we first convert the deriviation into a function which gives us all intermediate
      terms via function w.\<close>
    from 1(4)[unfolded relpow_fun_conv] obtain w
      where w: "w 0 = u" "w n = v" "(\<forall>i<n. (w i, w (Suc i)) \<in> estep)" by auto
    text \<open>And the proof now proceeds by case-analysis on whether any of these steps
      was a root step or whether all steps are non-root.\<close>
    show ?case sorry (* this will be the hardest proof, so don't start with this sorry *)
  qed
qed


text \<open>Next, completeness of CCA is easily established\<close>
lemma esteps_imp_CCA_st: assumes "CongS (CCA i) = CCA i" 
  shows "(s,t) \<in> estep^* \<longrightarrow> (s,t) \<in> CCA i" 
  sorry


subsection \<open>Soundness of CCA\<close>

text \<open>The crucial step to prove soundness is the following lemma, which
  might require some further auxiliary lemmas.\<close>
lemma Cong_esteps: "E \<subseteq> estep^* \<Longrightarrow> Cong E \<subseteq> estep^*" sorry (* this is the other sorry that is not trivial *)

text \<open>But you can easily verify that @{thm Cong_esteps} is the key to prove soundness of CCA.\<close>
lemma CCA_imp_esteps: "CCA i \<subseteq> estep^*" sorry


subsection \<open>Correctness of CCA\<close>

text \<open>Having soundness and completeness, correctness is simple.\<close>
theorem congruence_closure_correct: assumes "CongS (CCA i) = CCA i" 
  shows "(s,t) \<in> estep^* \<longleftrightarrow> (s, t) \<in> CCA i" 
  sorry

text \<open>The precondition @{term "CongS (CCA i) = CCA i"} can be discharged proving termination of 
  the congruence closure algorithm which just computes the least i such that the precondition is satisfied. 
  The existence of such an i follows from the fact that CCA i is increasing with increasing i and CCA i is bounded
  by the finite set of terms S x S. Proving termination formally is not part of this project.\<close>

end (* context with s t *)
end (* context with GE *)
end (* theory *)